import json
import threading
import os
from .distribute_params import run_scan
from .distribute_utils import return_current_machine_id, ssh_connect
from .distribute_utils import ssh_file_transfer, ssh_run, ssh_get_files
from .distribute_database import update_db


def run_central_machine(self, n_splits, run_central_node):
    '''Runs `talos.Scan()` in the central machine.

    Parameters
    ----------
    params  | `dict` | hyperparameter options

    Returns
    -------
    None.

    '''

    machine_id = 0

    run_scan(self, n_splits, run_central_node, machine_id)


def distribute_run(self):
    '''

    Parameters
    ----------
    run_central_machine | `bool` |The default is False.
    db_machine_id | `int` | The default is 0. Indicates the centralised store
                              where the data gets merged.

    Returns
    -------
    None.

    '''

    config = self.config_data

    if 'run_central_node' in config.keys():
        run_central_node = config['run_central_node']
    else:
        run_central_node = False

    update_db_n_seconds = 5
    if 'DB_UPDATE_INTERVAL' in config['database'].keys():
        update_db_n_seconds = int(config['database']['DB_UPDATE_INTERVAL'])

    n_splits = len(config['machines'])

    if run_central_node:
        n_splits += 1

    current_machine_id = str(return_current_machine_id(self))

    if current_machine_id == str(0):

        clients = ssh_connect(self)

        for machine_id, client in clients.items():

            new_config = config
            new_config['current_machine_id'] = machine_id

            with open('/tmp/remote_config.json', 'w') as outfile:
                json.dump(new_config, outfile)

            ssh_file_transfer(self, client, machine_id)

        # create the threads
        threads = []

        if run_central_node:

            args = (self, n_splits, run_central_node)
            thread = threading.Thread(target=run_central_machine, args=args)
            thread.start()
            threads.append(thread)

            args = ([self, update_db_n_seconds, current_machine_id, self.stage])
            thread = threading.Thread(target=update_db, args=args)
            thread.start()
            threads.append(thread)

        for machine_id, client in clients.items():

            args = (self, client, machine_id)
            thread = threading.Thread(target=ssh_run, args=args)
            thread.start()
            threads.append(thread)

        for t in threads:
            t.join()

        for file in os.listdir('/tmp/'):
            if file.startswith('machine_id'):
                os.remove('/tmp/' + file)

        for machine_id, client in clients.items():
            ssh_get_files(self, client, machine_id)

    from .distribute_finish import distribute_finish
    self = distribute_finish(self)



import json
import paramiko
import os
import pandas as pd
import datetime


def create_temp_file(self):
    filestr = '''
from jako import RemoteScan
import numpy as np
import json
import pickle

x=np.load('/tmp/x_data_remote.npy')
y=np.load('/tmp/y_data_remote.npy')

{}

with open('/tmp/arguments_remote.json','r') as f:
    arguments_dict=json.load(f)

t=RemoteScan(x=x,
             y=y,
             params=arguments_dict['params'],
             model={},
             experiment_name=arguments_dict['experiment_name'],
             x_val=arguments_dict['x_val'],
             y_val=arguments_dict['y_val'],
             val_split=arguments_dict['val_split'],
             random_method=arguments_dict['random_method'],
             seed=arguments_dict['seed'],
             performance_target=arguments_dict['performance_target'],
             fraction_limit=arguments_dict['fraction_limit'],
             round_limit=arguments_dict['round_limit'],
             time_limit=arguments_dict['time_limit'],
             boolean_limit=arguments_dict['boolean_limit'],
             reduction_method=arguments_dict['reduction_method'],
             reduction_interval=arguments_dict['reduction_interval'],
             reduction_window=arguments_dict['reduction_window'],
             reduction_threshold=arguments_dict['reduction_threshold'],
             reduction_metric=arguments_dict['reduction_metric'],
             minimize_loss=arguments_dict['minimize_loss'],
             disable_progress_bar=arguments_dict['disable_progress_bar'],
             print_params=arguments_dict['print_params'],
             clear_session=arguments_dict['clear_session'],
             save_weights=arguments_dict['save_weights'],
             config='/tmp/remote_config.json'
             )
    '''.format(self.model_func, self.model_name)

    with open("/tmp/scanfile_remote.py", "w") as f:
        f.write(filestr)


def return_current_machine_id(self,):
    ''' Return machine id after checking the ip from config'''

    current_machine_id = 0
    if 'current_machine_id' in self.config_data.keys():
        current_machine_id = int(self.config_data['current_machine_id'])

    return current_machine_id


def return_central_machine_id(self):
    ''' Return central machine id as mentioned in config'''
    central_id = 0
    config_data = self.config_data
    if 'database' in config_data.keys():
        central_id = int(config_data['database']['DB_HOST_MACHINE_ID'])
    return central_id


def read_config(self):
    '''Read config from file'''

    config_path = "/tmp/remote_config.json"

    with open(config_path, 'r') as f:
        config_data = json.load(f)

    return config_data


def write_config(self, new_config):
    ''' Write config to file'''

    config_path = "/tmp/remote_config.json"

    with open(config_path, 'w') as outfile:
        json.dump(new_config, outfile, indent=2)


def ssh_connect(self):
    '''
    Returns
    -------
    clients | `list` | List of client objects of machines after connection.

    '''

    configs = self.config_data['machines']
    clients = {}

    for config in configs:
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        host = config['JAKO_IP_ADDRESS']
        port = config['JAKO_PORT']
        username = config['JAKO_USER']

        if 'JAKO_PASSWORD' in config.keys():
            password = config['JAKO_PASSWORD']
            client.connect(host, port, username, password)

        elif 'JAKO_KEY_FILENAME' in config.keys():
            client.connect(host, port, username,
                           key_filename=config['JAKO_KEY_FILENAME'])

        clients[config['machine_id']] = client

    return clients


def ssh_file_transfer(self, client, machine_id):
    '''Transfer the current talos script to the remote machines'''

    create_temp_file(self)

    sftp = client.open_sftp()

    try:
        sftp.chdir(self.dest_dir)  # Test if dest dir exists

    except IOError:
        sftp.mkdir(self.dest_dir)  # Create dest dir
        sftp.chdir(self.dest_dir)

    data_files = ['y_data_remote.npy', 'x_data_remote.npy']
    scan_script_files = ['scanfile_remote.py']
    additional_scan_files = ['remote_config.json', 'arguments_remote.json']
    scan_filenames = data_files + scan_script_files + additional_scan_files
    for file in os.listdir("/tmp/"):
        if file in scan_filenames:
            sftp.put("/tmp/" + file, file)

    sftp.close()


def ssh_run(self, client, machine_id):
    '''Run the transmitted script remotely without args and show its output.

    Parameters
    ----------
    client | `Object` | paramiko ssh client object
    params | `dict`| hyperparameter options
    machine_id | `int`| Machine id for each of the distribution machines

    Returns
    -------
    None.

    '''
    execute_str = 'python3 /tmp/scanfile_remote.py'
    stdin, stdout, stderr = client.exec_command(execute_str)

    if stderr:
        for line in stderr:
            try:
                # Process each error line in the remote output
                print(line)
            except Exception as e:
                print(e)

    for line in stdout:
        try:
            # Process each line in the remote output
            print(line)
        except Exception as e:
            print(e)


def ssh_get_files(self, client, machine_id):
    '''Get files via ssh from a machine'''
    sftp = client.open_sftp()

    scan_object_filenames = ('scan_details.csv', 'scan_learning_entropy.csv',
                             'scan_round_history.npy', 'scan_round_times.csv',
                             'scan_saved_models.json', 'scan_saved_weights.npy',
                             'scan_data.csv')

    sftp.chdir(self.dest_dir)

    for file in sftp.listdir(self.dest_dir):
        if file.endswith(scan_object_filenames):
            sftp.get(self.dest_dir + file, '/tmp/' + file)

    sftp.close()


def fetch_latest_file(self):
    '''Fetch the latest csv for an experiment'''

    experiment_name = self.experiment_name
    save_timestamp = self.save_timestamp

    if not os.path.exists(experiment_name):
        return []

    filelist = []
    for file in os.listdir(experiment_name):
        if file.endswith('.csv'):
            if int(file.replace('.csv', '')) >= int(save_timestamp):
                latest_file = os.path.join(experiment_name, file)
                filelist.append(latest_file)

    if filelist:
        latest_filepath = max(filelist, key=os.path.getmtime)

        try:
            results_data = pd.read_csv(latest_filepath)
            return results_data
        except Exception as e:
            e = str(e)
            allowed_exception = 'No columns to parse from file'

            if allowed_exception in e:
                return []
            else:
                raise Exception(e)
    else:
        return []


def add_timestamp(self, results_data):
    '''Adds timestamp to the DataFrame'''

    ct = datetime.datetime.now()
    hour = ct.hour
    minute = ct.minute
    day = ct.day
    month = ct.month
    year = ct.year

    if minute < 10:
        minute = '0' + str(minute)

    if hour < 10:
        hour = '0' + str(hour)

    timestamp = "{}:{}/{}-{}-{}".format(hour, minute, day, month, year)
    results_data["timestamp"] = [timestamp] * len(results_data)

    return results_data


def get_experiment_stage(self, db):
    '''Get the current number of times of experiment run'''

    try:
        ids = db.return_existing_experiment_ids()
        stage = int(list(ids)[-1].split("-")[0]) + 1

    except Exception as e:
        allowed_exception = '(psycopg2.errors.UndefinedTable)'
        e = str(e)
        if allowed_exception in e:
            pass
        else:
            raise Exception(e)

        stage = 0

    return stage


def add_experiment_id(self, results_data, machine_id, start_row,
                      end_row, db, stage):
    '''Generate experiment id from model id and row number'''

    try:
        ids = db.return_existing_experiment_ids()
        if "experiment_id" in results_data.columns:
            results_data = results_data[~results_data['experiment_id'].isin(ids
                                                                            )]

    except Exception as e:
        allowed_exception = '(psycopg2.errors.UndefinedTable)'
        e = str(e)
        if allowed_exception in e:
            pass
        else:
            raise Exception(e)

    results_data = results_data.iloc[start_row:end_row]
    experiment_ids = []
    for i in range(start_row, end_row):
        experiment_id = str(stage) + "-" + str(machine_id) + "-" + str(i)
        experiment_ids.append(experiment_id)
    results_data["experiment_id"] = experiment_ids

    return results_data


def write_scan_namespace(self, scan_object, machine_id):
    '''

    Parameters
    ----------
    scan_object | talos.Scan object | Scan object after Scan run

    Returns
    -------
    None.

    '''
    import pandas as pd
    import json
    import numpy as np
    import os

    write_path = os.path.join('/tmp/', 'machine_id_' + str(machine_id) + '_')
    scan_details = scan_object.details
    scan_data = scan_object.data
    scan_learning_entropy = scan_object.learning_entropy
    scan_round_history = scan_object.round_history
    scan_round_times = scan_object.round_times
    scan_saved_models = scan_object.saved_models
    scan_saved_weights = scan_object.saved_weights

    details_df = pd.DataFrame({'scan_details': scan_details})
    details_df.to_csv(write_path + 'scan_details.csv')

    scan_data.to_csv(write_path + 'scan_data.csv')
    scan_learning_entropy.to_csv(write_path + 'scan_learning_entropy.csv')
    scan_round_times.to_csv(write_path + 'scan_round_times.csv')

    np.save(write_path + 'scan_round_history.npy', scan_round_history)

    with open(write_path + 'scan_saved_models.json', 'w') as f:
        scan_saved_models = {'saved_models_machine_id_' + str(machine_id):
                             scan_saved_models}
        json.dump(scan_saved_models, f, indent=2)

    np.save(write_path + 'scan_saved_weights.npy', scan_saved_weights)



from talos import Scan
from .distribute_utils import read_config, write_config, write_scan_namespace


def create_param_space(self, n_splits):
    '''Creates the parameter space for each machine

    Parameters
    ----------
    n_splits | int | The default is 2.

    Returns
    -------
    param_grid | `list` of `dict` | Split parameter spaces for each machine.

    '''

    from talos.parameters.DistributeParamSpace import DistributeParamSpace

    params = self.params
    param_keys = self.params.keys()

    random_method = self.random_method
    fraction_limit = self.fraction_limit
    round_limit = self.round_limit
    time_limit = self.time_limit
    boolean_limit = self.boolean_limit

    param_grid = DistributeParamSpace(params=params,
                                      param_keys=param_keys,
                                      random_method=random_method,
                                      fraction_limit=fraction_limit,
                                      round_limit=round_limit,
                                      time_limit=time_limit,
                                      boolean_limit=boolean_limit,
                                      machines=n_splits)

    return param_grid


def run_scan(self, machines, run_central_node, machine_id):
    '''Run `talos.Scan()` on each machine.

    Parameters
    ----------
    machines | int |
    run_central_node | bool |

    Returns
    -------
    None.

    '''

    # runs Scan in a machine after param split
    machine_id = int(machine_id)
    current_machine_id = machine_id

    if not run_central_node:
        if machine_id != 0:
            machine_id = machine_id - 1

    split_params = create_param_space(self, n_splits=machines)
    split_params = split_params.param_spaces[machine_id]

    print('Started experiment in machine id : {}'.format(current_machine_id))

    scan_object = Scan(x=self.x,
                       y=self.y,
                       params=split_params,
                       model=self.model,
                       experiment_name=self.experiment_name,
                       x_val=self.x_val,
                       y_val=self.y_val,
                       val_split=self.val_split,
                       random_method=self.random_method,
                       seed=self.seed,
                       performance_target=self.performance_target,
                       fraction_limit=self.fraction_limit,
                       round_limit=self.round_limit,
                       time_limit=self.time_limit,
                       boolean_limit=self.boolean_limit,
                       reduction_method=self.reduction_method,
                       reduction_interval=self.reduction_interval,
                       reduction_window=self.reduction_window,
                       reduction_threshold=self.reduction_threshold,
                       reduction_metric=self.reduction_metric,
                       minimize_loss=self.minimize_loss,
                       disable_progress_bar=self.disable_progress_bar,
                       print_params=self.print_params,
                       clear_session=self.clear_session,
                       save_weights=self.save_weights)

    print('Completed experiment in machine id : {}'.format(current_machine_id))

    new_config = read_config(self)
    new_config['finished_scan_run'] = True

    if int(current_machine_id) == 0:
        new_config['current_machine_id'] = 0

    write_config(self, new_config)
    write_scan_namespace(self, scan_object, current_machine_id)



from sqlalchemy import create_engine
from sqlalchemy_utils import database_exists, drop_database
from sqlalchemy.schema import DropTable


class Database:

    def __init__(self,
                 username=None,
                 password=None,
                 host=None,
                 port=None,
                 db_type='sqlite',
                 database_name='EXPERIMENT_LOG',
                 table_name='experiment_log',
                 encoding='LATIN1'):
        '''

        Parameters
        ----------
        username | str | The default is None.
        password | str | The default is None.
        host | str , optional| The default is None.
        port | str , optional| The default is None.
        db_type | str | The default is 'sqlite'.
        database_name | str | The default is 'EXPERIMENT_LOG'.
        table_name | str | The default is 'experiment_log'.

        Returns
        -------
        None.

        '''
        self.db_type = db_type
        self.database_name = database_name
        self.table_name = table_name
        self.encoding = encoding
        DB_URL = ''
        if db_type == 'sqlite':
            DB_URL = 'sqlite:///' + database_name + '.db'
        elif db_type == 'mysql':
            if port is None:
                port = 3306

            url = '''mysql+pymysql://{}:{}@{}:{}/{}'''.format(username,
                                                              password,
                                                              host,
                                                              str(port),
                                                              database_name)

            DB_URL = (url)

        elif db_type == 'postgres':
            if port is None:
                port = 5432

            url = 'postgresql://{}:{}@{}:{}/{}'.format(username,
                                                       password,
                                                       host,
                                                       str(port),
                                                       database_name)

            DB_URL = (url)

        self.DB_URL = DB_URL

    def create_db(self):
        '''
        Create database if it doesn't exists.
        '''

        engine = create_engine(self.DB_URL,
                               echo=False, isolation_level='AUTOCOMMIT')

        if not database_exists(engine.url):

            new_engine = create_engine(
                self.DB_URL.replace(self.database_name, ''),
                echo=False,
                isolation_level='AUTOCOMMIT',
            )
            conn = new_engine.connect()

            try:
                conn.execute(
                    '''
                    CREATE DATABASE {} ENCODING '{}'
                    '''.format(
                        self.database_name, self.encoding
                    )
                )

            except Exception as e:
                print(e)

        return engine

    def drop_db(self):
        '''
        Drop the database.
        '''
        drop_database(self.DB_URL)

    def delete_table(self, table_name):
        '''
        Delete the table.
        '''
        DropTable(table_name)

    def write_to_db(self, data_frame):
        '''


        Parameters
        ----------
        data_frame | `DataFrame` | DataFrame object consisting of tabular data.

        Returns
        -------
        None.

        '''
        engine = self.create_db()
        data_frame.to_sql(self.table_name, con=engine,
                          if_exists='append', index=False)

    def query_table(self, query):
        '''


        Parameters
        ----------
        query | `str`| Database query for the respective sql engine

        Returns
        -------
        res | `list` of `tuples` | Query output from the database

        '''
        engine = self.create_db()
        res = engine.execute(query).fetchall()
        return res

    def show_table_content(self):
        '''
        Returns
        -------
        res |`list` of `tuples` | Query output from the database

        '''
        res = self.query_table('SELECT * FROM {}'.format(self.table_name))
        return res

    def return_table_df(self):
        '''
        Returns
        -------
        data | Pandas DataFrame object | returns the database as a dataframe

        '''
        import pandas as pd

        table = self.show_table_content()
        data = pd.DataFrame(table)
        return data

    def return_existing_experiment_ids(self):
        '''
        Returns
        -------
        ids | Pandas Series object | returns the experiment id of the table

        '''

        query_str = 'SELECT experiment_id from {}'.format(self.table_name)
        res = self.query_table(query_str)
        res = [val[0] for val in res]

        return res

    def return_columns(self):
        '''
        Returns
        -------
        cols | list| returns the columns of the table
        '''

        query_string = """select COLUMN_NAME from information_schema.columns
                        where table_name='{}'"""
        query_string = query_string.format(self.table_name)
        cols = self.query_table(query_string)
        cols = [col[0] for col in cols]

        return cols

    def add_new_columns(self, columns):
        ''' Add a new column to the Database'''

        query_str = 'ALTER TABLE {}'.format(self.table_name)
        col_query_str = ' ADD COLUMN {} varchar,'

        for col in columns:
            query_str = query_str + col_query_str.format(col)

        query_str = query_str.rstrip(',') + ';'

        try:
            self.query_table(query_str)
        except Exception as e:
            exception_str1 = '''
                    This result object does not return rows.
                    '''
            exception_str2 = '(psycopg2.errors.DuplicateColumn)'
            exception_str3 = '(psycopg2.errors.UndefinedTable)'
            exceptions = [exception_str1, exception_str2, exception_str3]
            e = str(e)
            if any(ex in e for ex in exceptions):
                pass
            else:
                raise Exception(e)



from talos import Scan


class DistributedScan(Scan):

    def __init__(self,
                 x,
                 y,
                 params,
                 model,
                 experiment_name,
                 x_val=None,
                 y_val=None,
                 val_split=0.3,
                 random_method='uniform_mersenne',
                 seed=None,
                 performance_target=None,
                 fraction_limit=None,
                 round_limit=None,
                 time_limit=None,
                 boolean_limit=None,
                 reduction_method=None,
                 reduction_interval=50,
                 reduction_window=20,
                 reduction_threshold=0.2,
                 reduction_metric='val_acc',
                 minimize_loss=False,
                 disable_progress_bar=False,
                 print_params=False,
                 clear_session=True,
                 save_weights=True,
                 config='config.json'):
        '''Distributed version of talos.Scan() for the local machine.

        Parameters
        ----------
        params | `dict` | Hyperparameters for distribution.
        config | str or dict | The default is 'config.json'.

        Returns
        -------
        None.

        '''

        import time
        import json
        import os
        import inspect
        import numpy as np

        self.x = x
        self.y = y
        self.params = params
        self.model = model
        self.experiment_name = experiment_name
        self.x_val = x_val
        self.y_val = y_val
        self.val_split = val_split

        # randomness
        self.random_method = random_method
        self.seed = seed

        # limiters
        self.performance_target = performance_target
        self.fraction_limit = fraction_limit
        self.round_limit = round_limit
        self.time_limit = time_limit
        self.boolean_limit = boolean_limit

        # optimization
        self.reduction_method = reduction_method
        self.reduction_interval = reduction_interval
        self.reduction_window = reduction_window
        self.reduction_threshold = reduction_threshold
        self.reduction_metric = reduction_metric
        self.minimize_loss = minimize_loss

        # display
        self.disable_progress_bar = disable_progress_bar
        self.print_params = print_params

        # performance
        self.clear_session = clear_session
        self.save_weights = save_weights

        # distributed configurations
        self.config = config

        arguments_dict = self.__dict__
        remove_parameters = ["x", "y", "model"]
        arguments_dict = {k: v for k, v in arguments_dict.items()
                          if k not in remove_parameters}

        self.file_path = '/tmp/scanfile_remote.py'

        self.save_timestamp = time.strftime('%D%H%M%S').replace('/', '')

        # Handle the case when `config` is a filename
        if isinstance(config, str):
            with open(config, 'r') as f:
                self.config_data = json.load(f)

        # Handle the case when `config` is dict
        if isinstance(config, dict):
            self.config_data = config
            with open('config.json', 'w') as outfile:
                json.dump(self.config_data, outfile, indent=2)

        else:
            TypeError('`config` must be dict or filename string.')

        # write database name as same as experiment name
        self.config_data['database']['DB_TABLE_NAME'] = experiment_name

        if isinstance(config, str):
            config_path = config

        else:
            config_path = 'config.json'

        with open(config_path, 'w') as f:
            json.dump(self.config_data, f, indent=2)

        if 'finished_scan_run' in self.config_data.keys():
            del self.config_data['finished_scan_run']

        # handles location for params,data and model
        if not os.path.exists("/tmp/"):
            os.mkdir("/tmp/")

        self.dest_dir = "/tmp/"

        # save data in numpy format
        np.save("/tmp/x_data_remote.npy", x)
        np.save("/tmp/y_data_remote.npy", y)

        # get model function as a string
        model_func = inspect.getsource(model).lstrip()

        self.model_func = model_func
        self.model_name = model.__name__

        from .distribute_database import get_db_object
        from .distribute_utils import get_experiment_stage

        db = get_db_object(self)
        self.stage = get_experiment_stage(self, db)

        if not self.stage:
            self.stage = 0

        arguments_dict["stage"] = self.stage

        with open('/tmp/arguments_remote.json', 'w') as outfile:
            json.dump(arguments_dict, outfile, indent=2)

        with open('/tmp/remote_config.json', 'w') as outfile:
            json.dump(self.config, outfile, indent=2)

        from .distribute_run import distribute_run
        distribute_run(self)



def script_script():

    out = '''
from jako import RemoteScan
import numpy as np
import json
import pickle
x=np.load('/tmp/x_data_remote.npy')
y=np.load('/tmp/y_data_remote.npy')

{}
with open('/tmp/arguments_remote.json','r') as f:
    arguments_dict=json.load(f)

t=RemoteScan(x=x,
             y=y,
             params=arguments_dict['params'],
             model={},
             experiment_name=arguments_dict['experiment_name'],
             x_val=arguments_dict['x_val'],
             y_val=arguments_dict['y_val'],
             val_split=arguments_dict['val_split'],
             random_method=arguments_dict['random_method'],
             seed=arguments_dict['seed'],
             performance_target=arguments_dict['performance_target'],
             fraction_limit=arguments_dict['fraction_limit'],
             round_limit=arguments_dict['round_limit'],
             time_limit=arguments_dict['time_limit'],
             boolean_limit=arguments_dict['boolean_limit'],
             reduction_method=arguments_dict['reduction_method'],
             reduction_interval=arguments_dict['reduction_interval'],
             reduction_window=arguments_dict['reduction_window'],
             reduction_threshold=arguments_dict['reduction_threshold'],
             reduction_metric=arguments_dict['reduction_metric'],
             minimize_loss=arguments_dict['minimize_loss'],
             disable_progress_bar=arguments_dict['disable_progress_bar'],
             print_params=arguments_dict['print_params'],
             clear_session=arguments_dict['clear_session'],
             save_weights=arguments_dict['save_weights'],
             config='/tmp/remote_config.json'
             )'''

    return out



from talos import Scan


class RemoteScan(Scan):
    def __init__(self,
                 x,
                 y,
                 params,
                 model,
                 experiment_name,
                 **kwargs):
        '''Distributed version of talos.Scan() for the remote machines.

        Parameters
        ----------
        params | `dict` | Hyperparameters for distribution.
        config | str or dict | The default is '/tmp/remote_config.json'.

        Returns
        -------
        None.

        '''

        import time
        import json

        self.x = x
        self.y = y
        self.params = params
        self.model = model
        self.experiment_name = experiment_name
        self.x_val = kwargs['x_val']
        self.y_val = kwargs['y_val']
        self.val_split = kwargs['val_split']

        # randomness
        self.random_method = kwargs['random_method']
        self.seed = kwargs['seed']

        # limiters
        self.performance_target = kwargs['performance_target']
        self.fraction_limit = kwargs['fraction_limit']
        self.round_limit = kwargs['round_limit']
        self.time_limit = kwargs['time_limit']
        self.boolean_limit = kwargs['boolean_limit']

        # optimization
        self.reduction_method = kwargs['reduction_method']
        self.reduction_interval = kwargs['reduction_interval']
        self.reduction_window = kwargs['reduction_window']
        self.reduction_threshold = kwargs['reduction_threshold']
        self.reduction_metric = kwargs['reduction_metric']
        self.minimize_loss = kwargs['minimize_loss']

        # display
        self.disable_progress_bar = kwargs['disable_progress_bar']
        self.print_params = kwargs['print_params']

        # performance
        self.clear_session = kwargs['clear_session']
        self.save_weights = kwargs['save_weights']

        # distributed configurations
        self.config = kwargs['config']

        self.save_timestamp = time.strftime('%D%H%M%S').replace('/', '')

        if isinstance(self.config, str):
            with open(self.config, 'r') as f:
                self.config_data = json.load(f)

        else:
            TypeError('Pass the correct `config` path')

        if 'finished_scan_run' in self.config_data.keys():
            del self.config_data['finished_scan_run']

        config = self.config_data

        with open("/tmp/arguments_remote.json", "r") as f:
            arguments_dict = json.load(f)

        self.stage = arguments_dict["stage"]

        if 'run_central_node' in config.keys():
            run_central_node = config['run_central_node']
        else:
            run_central_node = False

        from .distribute_params import run_scan
        from .distribute_database import update_db
        from .distribute_utils import return_current_machine_id

        import threading

        n_splits = len(config['machines'])
        if run_central_node:
            n_splits += 1

        update_db_n_seconds = 5
        if 'DB_UPDATE_INTERVAL' in config['database'].keys():
            update_db_n_seconds = int(config['database']['DB_UPDATE_INTERVAL'])

        current_machine_id = str(return_current_machine_id(self))

        # create the threadpool
        threads = []

        args = ([self, update_db_n_seconds, current_machine_id, self.stage])
        thread = threading.Thread(target=update_db, args=args)
        thread.start()
        threads.append(thread)

        args = (self, n_splits, run_central_node, current_machine_id)
        thread = threading.Thread(target=run_scan, args=args)

        thread.start()
        threads.append(thread)

        # excecute the threadpool
        for t in threads:
            t.join()



def distribute_finish(self):
    import pandas as pd
    import os
    import json
    import numpy as np

    attrs_final = ['data', 'x', 'y', 'learning_entropy', 'round_times',
                   'params', 'saved_models', 'saved_weights', 'round_history',
                   'details']

    keys = list(self.__dict__.keys())
    for key in keys:
        if key not in attrs_final:
            delattr(self, key)

    from talos.scan.scan_addon import func_best_model, func_evaluate
    self.best_model = func_best_model.__get__(self)
    self.evaluate_models = func_evaluate.__get__(self)

    all_filenames = ['/tmp/' + file for file in os.listdir('/tmp/')]

    scan_data_list = []
    scan_details_list = []
    scan_learning_entropy_list = []
    scan_round_times_list = []
    scan_saved_models_dict = {}
    scan_round_history_dict = {}
    scan_saved_weights_dict = {}

    for file in all_filenames:
        if file.endswith('scan_data.csv'):
            df = pd.read_csv(file)
            scan_data_list.append(df)

        if file.endswith('scan_details.csv'):
            df = pd.read_csv(file)
            scan_details_list.append(df)

        if file.endswith('scan_learning_entropy.csv'):
            df = pd.read_csv(file)
            scan_learning_entropy_list.append(df)

        if file.endswith('scan_round_times.csv'):
            df = pd.read_csv(file)
            scan_round_times_list.append(df)

        if file.endswith('scan_saved_models.json'):
            with open(file, 'r') as f:
                scan_saved_model = json.load(f)
                scan_saved_models_dict.update(scan_saved_model)

        if file.endswith('scan_round_history.npy'):
            scan_round_history = np.load(file, allow_pickle=True)
            keyname = os.path.basename(file).replace(".npy", '')
            scan_round_history_dict[keyname] = scan_round_history

        if file.endswith('scan_saved_weights.npy'):
            scan_saved_weights = np.load(file, allow_pickle=True)
            keyname = os.path.basename(file).replace(".npy", '')
            scan_saved_weights_dict[keyname] = scan_saved_weights

    self.data = pd.concat(scan_data_list)
    self.details = pd.concat(scan_details_list)
    self.learning_entropy = pd.concat(scan_learning_entropy_list)
    self.round_times = pd.concat(scan_round_times_list)
    self.saved_models = scan_saved_models_dict
    self.round_history = scan_round_history_dict
    self.saved_weights = scan_saved_weights_dict

    return self



#! /usr/bin/env python
#
# Copyright (C) 2022 Autonomio

DESCRIPTION = 'Distributed Hyperparameter Experiments with Talos'
LONG_DESCRIPTION = '''\
Jako makes it straightforward to <strong>distribute Talos
experiments</strong> across one or more remote machines
without asking you to change anything in the way you are
already working with Talos.
'''

DISTNAME = 'jako'
MAINTAINER = 'Mikko Kotila'
MAINTAINER_EMAIL = 'mailme@mikkokotila.com'
URL = 'http://autonom.io'
LICENSE = 'MIT'
DOWNLOAD_URL = 'https://github.com/autonomio/jako/'
VERSION = '0.1.1'


try:
    from setuptools import setup

    _has_setuptools = True
except ImportError:
    from distutils.core import setup

install_requires = ['talos',
                    'numpy',
                    'pandas',
                    'paramiko',
                    'psycopg2-binary',
                    'sqlalchemy',
                    'sqlalchemy_utils']

if __name__ == '__main__':

    setup(
        name=DISTNAME,
        author=MAINTAINER,
        author_email=MAINTAINER_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        license=LICENSE,
        url=URL,
        version=VERSION,
        download_url=DOWNLOAD_URL,
        install_requires=install_requires,
        packages=[
            'jako',
            'jako.distribute',
            'jako.database',
        ],
        classifiers=[
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'License :: OSI Approved :: MIT License',
            'Topic :: Scientific/Engineering :: Human Machine Interfaces',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'Topic :: Scientific/Engineering :: Mathematics',
            'Operating System :: POSIX',
            'Operating System :: Unix',
            'Operating System :: MacOS',
            'Operating System :: Microsoft :: Windows :: Windows 10',
        ],
    )



import time
from .distribute_utils import read_config, write_config
from .distribute_utils import add_experiment_id, add_timestamp
from .distribute_utils import fetch_latest_file
import sys


def get_db_object(self):
    config = self.config_data
    from ..database.database import Database

    machine_config = config['machines']
    db_config = config['database']
    username = db_config['DB_USERNAME']
    password = db_config['DB_PASSWORD']

    host_machine_id = int(db_config['DB_HOST_MACHINE_ID'])

    for machine in machine_config:
        if int(machine['machine_id']) == host_machine_id:
            host = machine['JAKO_IP_ADDRESS']
            break

    port = db_config['DB_PORT']
    database_name = db_config['DATABASE_NAME']
    db_type = db_config['DB_TYPE']
    table_name = db_config['DB_TABLE_NAME']
    encoding = db_config['DB_ENCODING']

    db = Database(username,
                  password,
                  host,
                  port,
                  database_name=database_name,
                  db_type=db_type,
                  table_name=table_name,
                  encoding=encoding)

    return db


def update_db(self, update_db_n_seconds, current_machine_id, stage):
    '''Make changes to the datastore based on a time interval

    Parameters
    ----------
    update_db_n_seconds | int | Time interval required to update the db

    Returns
    -------
    db | Database object | Database object with engine

    '''

    # update the database every n seconds
    db = get_db_object(self)
    config = self.config_data

    def __start_upload(results_data):

        if len(results_data) > 0:
            db_cols = db.return_columns()
            df_cols = results_data.columns
            missing_columns = [col for col in db_cols if col not in df_cols]
            new_columns = [col for col in df_cols if col not in db_cols]

            if len(missing_columns) > 0:
                exception_str = '''You have to change the experiment_name or
                add at least value for {}
                into the input parameter'''.format(missing_columns)
                raise Exception(exception_str)

            if len(new_columns) > 0:
                db.add_new_columns(new_columns)

            db.write_to_db(results_data)
        return db

    start_time = int(self.save_timestamp)

    start_row = 0
    end_row = 0

    while True:

        new_time = int(time.strftime('%D%H%M%S').replace('/', ''))

        if new_time - start_time >= update_db_n_seconds:

            if 'database' in config.keys():

                results_data = fetch_latest_file(self)

                if len(results_data) == 0:

                    start_time = new_time
                    time.sleep(update_db_n_seconds)
                    continue

                if len(results_data) > 0:
                    start_row = end_row
                    end_row = len(results_data)

                    if start_row != end_row and end_row > start_row:
                        results_data = add_timestamp(self, results_data)
                        results_data = add_experiment_id(self,
                                                         results_data,
                                                         current_machine_id,
                                                         start_row,
                                                         end_row,
                                                         db,
                                                         stage)

                        __start_upload(results_data)

                new_config = read_config(self)

                if 'finished_scan_run' in new_config.keys():

                    results_data = fetch_latest_file(self)

                    start_row = end_row
                    end_row = len(results_data)

                    if start_row != end_row and end_row > start_row:
                        results_data = add_timestamp(self, results_data)
                        results_data = add_experiment_id(self,
                                                         results_data,
                                                         current_machine_id,
                                                         start_row,
                                                         end_row,
                                                         db,
                                                         stage)

                        __start_upload(results_data)
                        write_config(self, new_config)

                    sys.exit()

                else:

                    start_time = new_time
                    time.sleep(update_db_n_seconds)

            else:
                print('Database credentials not given.')
